/*
 * amplitude_adc.c
 *
 *  Created on: Jul 23, 2016
 *      Author: esp8266
 */

#include <stdlib.h>
#include <avr/io.h>
#include <amplitude_adc.h>

#define AMP_HIST_BUFF_SIZE 8
#define AMP_NOISE_TRESHOLD 8
#define AMP_MAX_AVG_DEVIATION 20
#define AMP_DEVIATION_FACTOR 2
#define AMP_MAX_RETRIES 30

static uint16_t amp_sample_history[AMP_HIST_BUFF_SIZE];

static uint16_t amp_samples_sum;

static uint16_t amp_samples_avg;

static uint8_t amp_sample_history_pos;

void amp_init()
{

	ADMUX = (1 << REFS1) //
	| (1 << REFS0) //
			| (0 << ADLAR) //
			| (0 << MUX4) //
			| (0 << MUX3) //
			| (0 << MUX2) //
			| (0 << MUX1) //
			| (0 << MUX0) //
			;

	ADCSRA = (1 << ADEN) // AD enable
	| (0 << ADSC) // AD start conversion
			| (0 << ADATE) // AD auto trigger enable
			| (0 << ADIF) // AD interrupt flag
			| (0 << ADIE) // AD interrupt enable
			| (0 << ADPS2) // AD prescaler select
			| (1 << ADPS1) // Prescaler to 1:8
			| (1 << ADPS0) // 125kHz @ F_CPU == 1000000
			;

}

static inline uint8_t amp_conversion_complete()
{
	return (ADCSRA & (1 << ADIF));
}

static inline uint16_t amp_get_sample()
{
	// Start conversion
	ADCSRA |= (1 << ADSC);

	// meanwhile remove the sample to be overridden from history
	amp_samples_sum -= amp_sample_history[amp_sample_history_pos];

	// wait for conversion to complete
	while (!amp_conversion_complete())
	{
		// wait
	}

	uint16_t sample = ADC;

	amp_sample_history[amp_sample_history_pos] = sample;
	amp_samples_sum += sample;

	amp_sample_history_pos ++;
	amp_sample_history_pos %= AMP_HIST_BUFF_SIZE;


	return sample;
}

static inline
void amp_fill_sample_history()
{
	for (int i = 0; i < AMP_HIST_BUFF_SIZE; ++i) {
		amp_get_sample();
	}
	amp_samples_avg = amp_samples_sum / AMP_HIST_BUFF_SIZE;
}

static
uint16_t amp_calc_avg_deviation(){
	uint16_t deviation = 0;
	for (int i = 0; i < AMP_HIST_BUFF_SIZE; ++i) {
		deviation += abs(amp_samples_avg - amp_sample_history[i]);
	}

	deviation /= AMP_HIST_BUFF_SIZE;
	return deviation;
}

static
uint8_t amp_samples_too_divergent(){
	uint16_t avg_dev = amp_calc_avg_deviation();

	if (avg_dev > AMP_MAX_AVG_DEVIATION){
		return 1;
	}

	for (int i = 0; i < AMP_HIST_BUFF_SIZE; ++i) {
		uint16_t dev =  abs(amp_samples_avg - amp_sample_history[i]);
		if ((dev > AMP_NOISE_TRESHOLD) && (dev > avg_dev * AMP_DEVIATION_FACTOR)){
			return 1;
		}
	}

	return 0;
}

uint16_t amp_get_amplitude()
{
	amp_fill_sample_history();

	uint8_t retries = 0;
	while ((retries < AMP_MAX_RETRIES) && amp_samples_too_divergent()){
		amp_get_sample();
		retries ++;
	}

	return amp_samples_avg;
}
