/*
 * scanner.c
 *
 *  Created on: Jul 24, 2016
 *      Author: esp8266
 */

#include <stdint.h>
#include <string.h>
#include "oscillator.h"
#include "amplitude_adc.h"
#include "mode.h"
#include "scanner.h"

#define SCAN_START_PERIOD 3000
#define SCAN_END_PERIOD 1100

void (*scan_on_period_scanned)(const mode_t *p_mode_candidate) = 0;
void (*scan_on_mode_detected)(const mode_t *p_modes, uint8_t count) = 0;
void (*scan_on_chase_change)(const uint16_t period, const uint16_t min_period,
		const uint16_t max_period);

#define SCAN_MODES_COUNT 3
#define SCAN_MIN_MODE_PERIOD_DISTANCE 100
#define SCAN_MODE_HEIGHT 15

// modes that have been detected by the last run of scan_scan.
static
mode_t scan_modes[SCAN_MODES_COUNT];

// A structure to enclose data for scan_scan and the functions called by scan_scan exclusively.
struct scan_context
{
	// Represents the mode candidate of the previous iteration within scan_scan.
	// prev_cand.period is one bigger than the period currently investigated.
	// prev_cand.amplitude has been measured already and is valid.
	// prev_cand.width is invalid, initialized with 0xff and has not yet been updated.
	mode_t prev_cand;

	// does the amplitude rise with increasing frequency?
	uint8_t amplitude_was_rising;
};

// Data context of the function scan_scan. Only valid during
// execution of scan_scan.
static
struct scan_context scan_ctx;

static inline
void scan_clear_modes()
{
	memset(scan_modes, 0, sizeof scan_modes);
//	scan_modes[0].period = 10;
//	scan_modes[0].amplitude = 11;
//
//	scan_modes[1].period = 20;
//	scan_modes[1].amplitude = 21;
//
//	scan_modes[2].period = 30;
//	scan_modes[2].amplitude = 31;

//	scan_on_mode_detected(scan_modes, SCAN_MODES_COUNT);
}

static
void scan_sort_modes()
{
	uint8_t swapped;
	do
	{
		swapped = 0;
		for (int i = 0; i < SCAN_MODES_COUNT - 1; ++i)
		{
			if (scan_modes[i].period > scan_modes[i + 1].period)
			{
				mode_t temp;
				memcpy(&temp, &scan_modes[i], sizeof(mode_t));
				memcpy(&scan_modes[i], &scan_modes[i + 1], sizeof(mode_t));
				memcpy(&scan_modes[i + 1], &temp, sizeof(mode_t));

				swapped++;
			}
		}

	} while (swapped);
}

//// comparer for scan_find_mode. Used to find the mode with the lowest amplitude.
//static uint8_t scan_lower_amplitude_comparer(const mode_t* p_left,
//		const mode_t* p_right)
//{
//	return p_left->amplitude < p_right->amplitude;
//}

// comparer for scan_find_mode. Used to find the mode with the nearest period to
static uint8_t scan_closer_period_comparer(const mode_t* p_left,
		const mode_t* p_right)
{
	// because the the periods are swept through decreasingly, we can rely on
	// scan_ctx.prev_mode_candidate.period to be lesser then the periods of previously found
	// modes.
	uint16_t left_distance = p_left->period
			- scan_ctx.prev_cand.period;
	uint16_t right_distance = p_right->period
			- scan_ctx.prev_cand.period;

	return left_distance < right_distance;
}

static uint8_t scan_wide_mode_comparer(const mode_t* p_left,
		const mode_t* p_right)
{
	if (p_right->width == 0)
	{
		return 0;
	}
	if (p_left->width == 0)
	{
		return 1;
	}
	return p_left->width > p_right->width;
}

static uint8_t scan_narrow_mode_comparer(const mode_t* p_left,
		const mode_t* p_right)
{
	return p_left->width < p_right->width;
}

static inline mode_t* scan_find_mode(
		uint8_t (*comparer)(const mode_t* p_left, const mode_t* p_right))
{
	mode_t * p_beast_mode = &scan_modes[0];
	for (int i = 1; i < SCAN_MODES_COUNT; ++i)
	{
		if (comparer(&scan_modes[i], p_beast_mode))
		{
			p_beast_mode = &scan_modes[i];
		}
	}
	return p_beast_mode;
}

static inline
void scan_update_last_mode_width(mode_t *p_mode_to_update)
{
	uint8_t period_delta = p_mode_to_update->period
			- scan_ctx.prev_cand.period;

	// Are we horizontally (period) in a range where we might want to update the width?
	// Once the width (initialized with 0xff) was updated this condition also takes care that it won't be re-updated.
	if (period_delta < p_mode_to_update->width)
	{
		// Are we vertically (amplitude) in a range where we might want to update the width?
		if (p_mode_to_update->amplitude
				> scan_ctx.prev_cand.amplitude + SCAN_MODE_HEIGHT)
		{

			p_mode_to_update->width = p_mode_to_update->period
					- scan_ctx.prev_cand.period;
		}
	}
}

static inline
void scan_investigate_prev_mode_candidate()
{
	// We want to prevent to create clusters of mode entries
	// which differ only marginally by period and effectively represent the same
	// mode. Thus we want to make sure that the period differences exceed a
	// minimum delta.

	// Let's start by investigating the mode with the most similar period.
	mode_t * p_replace_mode = scan_find_mode(scan_closer_period_comparer);

	scan_update_last_mode_width(p_replace_mode);

	// Does the period of the candidate differ enough from previously found modes
	// to justify the creation of a new one?
	if (p_replace_mode->period
			- scan_ctx.prev_cand.period> SCAN_MIN_MODE_PERIOD_DISTANCE)
	{
		// The period differs enough.
		// -> don't update the mode entry for a period far away
		// -> use the weakest found mode as a candidate to replace
		p_replace_mode = scan_find_mode(scan_wide_mode_comparer);
	}

	// only replace the mode if it really is a stronger mode
	if (p_replace_mode->amplitude < scan_ctx.prev_cand.amplitude)
	{
		memcpy(p_replace_mode, &scan_ctx.prev_cand, sizeof(mode_t));
		scan_sort_modes();
	}
}

static inline
void scan_clear_context()
{
	scan_ctx.prev_cand.period = SCAN_START_PERIOD;
	scan_ctx.prev_cand.amplitude = 0;
	scan_ctx.amplitude_was_rising = 1;
}

static inline
void scan_detect_mode(const uint16_t amplitude)
{
	// When an increase of frequency results in a drop of amplitude we just
	// have passed a mode.
	if (scan_ctx.amplitude_was_rising)
	{
		if (amplitude < scan_ctx.prev_cand.amplitude)
		{
			// we have found a new mode candidate
			scan_investigate_prev_mode_candidate();

			// notify listener
			if (scan_on_mode_detected)
			{
				scan_on_mode_detected(scan_modes, SCAN_MODES_COUNT);
			}
		}
	}

	scan_ctx.amplitude_was_rising = (amplitude
			>= scan_ctx.prev_cand.amplitude);
}

void scan_scan()
{
	scan_clear_context();
	scan_clear_modes();
	osc_enable(1);

	for (uint16_t period = SCAN_START_PERIOD; period > SCAN_END_PERIOD;
			period--)
	{
		osc_set_period(period);

		uint16_t amplitude = amp_get_amplitude();

		scan_detect_mode(amplitude);

		scan_ctx.prev_cand.period = period;
		scan_ctx.prev_cand.amplitude = amplitude;
		scan_ctx.prev_cand.width = 0xFF;

		// notify listener
		if (scan_on_period_scanned)
		{
			scan_on_period_scanned(&scan_ctx.prev_cand);
		}
	}
	osc_enable(0);
}

void scan_chase()
{
	uint16_t period = 500;

	uint16_t min_period;
	uint16_t max_period;
	int8_t dir = -1;
	uint16_t max_amplitude = SCAN_MODE_HEIGHT + 1;

	uint8_t mode_to_wide_count;
	while (1)
	{
		scan_scan();
		mode_t *p_mode = scan_find_mode(scan_narrow_mode_comparer);
		period = p_mode->period;
		min_period = period;
		max_period = period;
		mode_to_wide_count = 0;
		osc_enable(1);
		while (mode_to_wide_count < 5)
		{
			osc_set_period(period);

			uint16_t amplitude = amp_get_amplitude();

			if (amplitude < max_amplitude - SCAN_MODE_HEIGHT)
			{
				if (dir > 0)
				{
					max_period = period;
				}
				else
				{
					min_period = period;
				}
				if (max_period - min_period > SCAN_MIN_MODE_PERIOD_DISTANCE)
				{
					mode_to_wide_count++;
				}
				else
				{
					if (mode_to_wide_count > 0)
					{
						mode_to_wide_count--;
					}
				}
				dir *= -1;
				max_amplitude = amplitude;
			}

//		if (max_period < period) {max_period = period;}
//		if (min_period > period) {min_period = period;}
			if (max_amplitude < amplitude)
			{
				max_amplitude = amplitude;
			}
			period += dir;
			if (scan_on_chase_change)
			{
				scan_on_chase_change(period, min_period, max_period);
			}
		}
	}
}
